/*
 * libdevice-node
 *
 * Copyright (c) 2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <errno.h>
#include <string.h>
#include <stdbool.h>

#include <libsyscommon/dbus-systemd.h>

#include <hw/shared.h>
#include <hw/usb_client.h>

#define MAX_GADGET_STR_LEN 256

#define LEGACY_ROOTPATH        "/sys/class/usb_mode/usb0"

/* Device descriptor values */
#define LEGACY_ID_VENDOR_PATH       LEGACY_ROOTPATH"/idVendor"
#define LEGACY_ID_PRODUCT_PATH      LEGACY_ROOTPATH"/idProduct"
#define LEGACY_BCD_DEVICE_PATH      LEGACY_ROOTPATH"/bcdDevice"
#define LEGACY_CLASS_PATH           LEGACY_ROOTPATH"/bDeviceClass"
#define LEGACY_SUBCLASS_PATH        LEGACY_ROOTPATH"/bDeviceSubClass"
#define LEGACY_PROTOCOL_PATH        LEGACY_ROOTPATH"/bDeviceProtocol"

/* Strings */
#define LEGACY_IMANUFACTURER_PATH   LEGACY_ROOTPATH"/iManufacturer"
#define LEGACY_IPRODUCT_PATH        LEGACY_ROOTPATH"/iProduct"

/* Functions in each config */
#define LEGACY_CONFIG_1_PATH        LEGACY_ROOTPATH"/funcs_fconf"
#define LEGACY_CONFIG_2_PATH        LEGACY_ROOTPATH"/funcs_sconf"

/* should be single char */
#define LEGACY_FUNC_SEP             ","

/* ON/OFF switch */
#define LEGACY_ENABLE_PATH          LEGACY_ROOTPATH"/enable"
#define LEGACY_ENABLE               "1"
#define LEGACY_DISABLE              "0"

#ifndef EXPORT
#define EXPORT	__attribute__ ((visibility("default")))
#endif

static struct usb_function *legacy_find_func(const char *name)
{
	int i;

	for (i = 0; _available_funcs[i]; ++i)
		if (!strcmp(name, _available_funcs[i]->name))
			return _available_funcs[i];

	return NULL;
}


static bool legacy_is_function_supported(struct usb_client *usb,
					 struct usb_function *func)
{
	/*
	 * TODO
	 * Instead of only checking whether we know this function
	 * we should also parse sysfs to check if it is build into
	 * slp-gadget.
	 */
	if (legacy_find_func(func->name))
		return true;
	else
		return false;
}

static bool legacy_is_gadget_supported(struct usb_client *usb,
				       struct usb_gadget *gadget)
{
	int i, j;

	if (!gadget || !gadget->configs || !gadget->funcs)
		return false;

	/*
	 * TODO
	 * Here is a good place to ensure that serial is immutable
	 */
	if (gadget->strs) {
		/* only strings in US_en are allowed */
		if (gadget->strs[0].lang_code != 0x409 ||
		    gadget->strs[1].lang_code)
			return false;
	}

	for (j = 0; gadget->configs[j]; ++j) {
		struct usb_configuration *config = gadget->configs[j];

		if (config->strs && config->strs[0].lang_code)
			return false;

		if (!config->funcs)
			return false;

		for (i = 0; config->funcs[i]; ++i)
			if (!legacy_is_function_supported(usb, config->funcs[i]))
				return false;
	}

	if (j == 0 || j > 2)
		return false;

	return true;
}

/* TODO. Maybe move this to sys ? */
static int legacy_set_int_hex(char *path, int val)
{
	char buf[MAX_GADGET_STR_LEN];
	int r;

	if (!path)
		return -EINVAL;

	snprintf(buf, sizeof(buf), "%x", val);
	r = sys_set_str(path, buf);
	if (r < 0)
		return r;

	return 0;
}

static int legacy_set_gadget_attrs(struct usb_gadget_attrs *attrs)
{
	int ret;

	ret = sys_set_int(LEGACY_CLASS_PATH, attrs->bDeviceClass);
	if (ret)
		return ret;

	ret = sys_set_int(LEGACY_SUBCLASS_PATH, attrs->bDeviceSubClass);
	if (ret)
		return ret;

	ret = sys_set_int(LEGACY_PROTOCOL_PATH, attrs->bDeviceProtocol);
	if (ret)
		return ret;

	ret = legacy_set_int_hex(LEGACY_ID_VENDOR_PATH, attrs->idVendor);
	if (ret)
		return ret;

	ret = legacy_set_int_hex(LEGACY_ID_PRODUCT_PATH, attrs->idProduct);
	if (ret)
		return ret;

	ret = legacy_set_int_hex(LEGACY_BCD_DEVICE_PATH, attrs->bcdDevice);

	return ret;
}

static int legacy_set_gadget_strs(struct usb_gadget_strings *strs)
{
	int ret = 0;

	/*
	 * TODO
	 * Here is a good place to ensure that serial is immutable
	 */

	if (strs->manufacturer) {
		ret = sys_set_str(LEGACY_IMANUFACTURER_PATH, strs->manufacturer);
		if (ret)
			return ret;
	}

	if (strs->product) {
		ret = sys_set_str(LEGACY_IPRODUCT_PATH, strs->product);
		if (ret)
			return ret;
	}

	return ret;
}

static int legacy_set_gadget_config(char *cpath,
				    struct usb_configuration *config)
{
	char buf[MAX_GADGET_STR_LEN];
	int left = sizeof(buf);
	char *pos = buf;
	int ret;
	int i;

	if (!config) {
		buf[0] = '\n';
		buf[1] = '\0';
		goto empty_config;
	}

	for (i = 0; config->funcs[i]; ++i) {
		ret = snprintf(pos, left, "%s" LEGACY_FUNC_SEP,
			       config->funcs[i]->name);
		if (ret >= left)
			return -EOVERFLOW;

		pos += ret;
		left -= ret;
	}

	/* eliminate last separator */
	*(pos - 1) = '\0';

empty_config:
	return sys_set_str(cpath, buf);
}

static int legacy_reconfigure_gadget(struct usb_client *usb,
				     struct usb_gadget *gadget)
{
	int ret;

	if (!usb || !gadget || !legacy_is_gadget_supported(usb, gadget))
		return -EINVAL;

	ret = legacy_set_gadget_attrs(&gadget->attrs);
	if (ret)
		return ret;

	if (gadget->strs) {
		ret = legacy_set_gadget_strs(gadget->strs + 0);
		if (ret)
			return ret;
	}

	ret = legacy_set_gadget_config(LEGACY_CONFIG_1_PATH, gadget->configs[0]);
	if (ret)
		return ret;

	ret = legacy_set_gadget_config(LEGACY_CONFIG_2_PATH, gadget->configs[1]);

	return ret;
}

static void legacy_start_stop_service_and_handler(bool start)
{
	int i;
	int ret;
	int n_func = 0;
	char *ptr;
	char *begin;
	char *fname;
	char *sep = LEGACY_FUNC_SEP;
	char buf[MAX_GADGET_STR_LEN];
	struct usb_function *func;
	struct usb_function *funcs[USB_FUNCTION_IDX_MAX];

	/* SLP gadget uses two USB configuration.
	 * (/sys/class/usb_mode/usb0/funcs_fconf and /sys/class/usb_mode/usb0/funcs_sconf)
	 *
	 * One usb function can be included in two configurations simultaneously.
	 * In this situation, a handler associated with function can be called twice for a usb function.
	 * To prevent duplicate calls,
	 * Collect all functions and remove duplicates and process them.
	 */

	/* First configuration */
	ret = sys_get_str(LEGACY_CONFIG_1_PATH, buf, sizeof(buf));
	if (ret)
		goto second_configuration;

	/* Caution: buf ends with '\n' */
	ptr = strchr(buf, '\n');
	if (ptr)
		*ptr = 0;

	begin = buf;
	for (fname = strsep(&begin, sep); fname; fname = strsep(&begin, sep)) {
		func = legacy_find_func(fname);
		if (!func)
			continue;

		for (i = 0; i < n_func; i++)
			if (funcs[i] == func)
				continue;

		if(n_func >= USB_FUNCTION_IDX_MAX) /* What happen */
			break;

		funcs[n_func] = func;
		n_func++;
	}

	/* Second configuration */
second_configuration:
	ret = sys_get_str(LEGACY_CONFIG_2_PATH, buf, sizeof(buf));
	if (ret)
		return;

	/* Caution: buf ends with '\n' */
	ptr = strchr(buf, '\n');
	if (ptr)
		*ptr = 0;

	begin = buf;
	for (fname = strsep(&begin, sep); fname; fname = strsep(&begin, sep)) {
		func = legacy_find_func(fname);
		if (!func)
			continue;

		for (i = 0; i < n_func; i++)
			if (funcs[i] == func)
				continue;

		if(n_func >= USB_FUNCTION_IDX_MAX) /* What happen */
			break;

		funcs[n_func] = func;
		n_func++;
	}

	for (i = 0; i < n_func; i++) {
		if (start) {
			if (funcs[i]->handler)
				funcs[i]->handler(1);

			if (funcs[i]->service)
				(void)systemd_start_unit_wait_started(funcs[i]->service, ".service", -1);
		} else {
			if (funcs[i]->service)
				(void)systemd_stop_unit_wait_stopped(funcs[i]->service, ".service", -1);

			if (funcs[i]->handler)
				funcs[i]->handler(0);
		}
	}
}

static int legacy_enable(struct usb_client *usb)
{
	int ret;

	ret = sys_set_str(LEGACY_ENABLE_PATH, LEGACY_ENABLE);
	if (ret < 0)
		return ret;

	legacy_start_stop_service_and_handler(true);

	return 0;
}

static int legacy_disable(struct usb_client *usb)
{

	legacy_start_stop_service_and_handler(false);

	return sys_set_str(LEGACY_ENABLE_PATH, LEGACY_DISABLE);
}

EXPORT
int hw_legacy_gadget_open(struct hw_info *info,
		const char *id, struct hw_common **common)
{
	struct usb_client *legacy;

	if (!info || !common)
		return -EINVAL;

	/* check if slp usb gadget exists */
	if (access("/sys/class/usb_mode/usb0/enable", F_OK))
		return -ENOENT;

	legacy = calloc(1, sizeof(*legacy));
	if (!legacy)
		return -ENOMEM;

	legacy->common.info = info;

	legacy->reconfigure_gadget = legacy_reconfigure_gadget;
	legacy->enable = legacy_enable;
	legacy->disable = legacy_disable;

	*common = &legacy->common;
	return 0;
}

EXPORT
int hw_legacy_gadget_close(struct hw_common *common)
{
	struct usb_client *legacy;

	if (!common)
		return -EINVAL;

	legacy = container_of(common, struct usb_client,
					 common);

	free(legacy);
	return 0;
}
